
import torchvision.transforms as transforms
import torch.nn as nn
import seaborn as sn
import pandas as pd
import numpy as np
import torchvision
import datetime
import sklearn
import torch
import os
from tqdm.auto import  tqdm,trange
from pytorch_metric_learning import distances, losses, miners, reducers, testers
from torch.utils.data import Dataset, DataLoader
from torchvision.datasets import CIFAR10
from tensorboardX import SummaryWriter
from matplotlib.collections import QuadMesh
import matplotlib.font_manager as fm
import matplotlib.pyplot as plt

##################################################################################################################################################

class ConvBlock(nn.Module):
    def __init__(self,in_channels,layer_map,in_shape=None,num_classes=10):
        super(ConvBlock,self).__init__()
        self.use_fc = False
        layers = []
        for layer in layer_map:
            if type(layer)==int:
                layers.append(nn.Conv2d(in_channels=in_channels
                                                     ,out_channels=layer,kernel_size=3,padding=1))
                layers.append(nn.BatchNorm2d(layer))
                layers.append(nn.ReLU())
                
            else:
                layers.append(nn.MaxPool2d(2))
        self.conv_layers = nn.Sequential(*layers)
        self.pool = nn.AvgPool2d(kernel_size=1, stride=1)
        if in_shape!=None:
            self.use_fc = True
            self.fcs = nn.Sequential(nn.Flatten(start_dim=1),
                                    nn.Linear(self.get_dim(in_shape),512),nn.ReLU(),
                                    nn.Linear(512,num_classes))
    def get_dim(self,in_shape):
        X = torch.rand([1,*in_shape])
        return self.pool(self.conv_layers(X)).flatten(1).shape[-1]
    
    def forward(self,X):
        if self.use_fc:
            X = self.conv_layers(X)
            X = self.pool(X)
            X = self.fcs(X)
            return X
         
        else:
            return self.conv_layers(X)
            
##################################################################################################################################################

class ModuleAdder(nn.Module):
    def __init__(self,module1,module2,in_shape=[3,32,32],num_classes=10,freeze=True):
        super(ModuleAdder,self).__init__()
        if freeze:
            for p in module1.parameters():
                p.requires_grad = False
            
        self.conv_layers = nn.Sequential(module1,module2)
        self.pool = nn.AvgPool2d(kernel_size=1, stride=1)
        self.fcs = nn.Sequential(nn.Flatten(start_dim=1),
                                    nn.Linear(self.get_dim(in_shape),512),nn.ReLU(),
                                    nn.Linear(512,num_classes))
      
    
    def get_dim(self,in_shape):
        X = torch.rand([1,*in_shape])
        return self.pool(self.conv_layers(X)).flatten(1).shape[-1]
    def forward(self,X):
        X = self.conv_layers(X)
        X = self.pool(X)
        X = self.fcs(X)
        return X

##################################################################################################################################################

class MLP(nn.Module):
    def __init__(self,Filter):
        super(MLP,self).__init__()
        self.filter = Filter.cpu()
        for p in self.filter.parameters():
                p.requires_grad = True
        X = torch.rand(10,3,32,32)
        dim = self.filter(X).flatten(start_dim=1).shape[-1]
        self.fc = nn.Sequential(nn.Flatten(start_dim=1),nn.Linear(dim,512),nn.ReLU(),nn.Linear(dim,10))
    def forward(self, X):
        return self.fc(self.filter(X))

##################################################################################################################

def accuracy(out,y):
    with torch.no_grad():
        return ((torch.argmax(out,dim=1) == y).sum()/len(y)).item()

##################################################################################################################

def get_new_fig(fn, figsize=[9, 9]):
    fig1 = plt.figure(fn, figsize)
    ax1 = fig1.gca()  # Get Current Axis
    ax1.cla()  # clear existing plot
    return fig1, ax1

##################################################################################################################

def configcell_text_and_colors( array_df, lin, col, oText, facecolors, posi, fz, fmt, show_null_values=0):
    text_add = []
    text_del = []
    cell_val = array_df[lin][col]
    tot_all = array_df[-1][-1]
    per = (float(cell_val) / tot_all) * 100
    curr_column = array_df[:, col]
    ccl = len(curr_column)

    # last line  and/or last column
    if (col == (ccl - 1)) or (lin == (ccl - 1)):
        # tots and percents
        if cell_val != 0:
            if (col == ccl - 1) and (lin == ccl - 1):
                tot_rig = 0
                for i in range(array_df.shape[0] - 1):
                    tot_rig += array_df[i][i]
                per_ok = (float(tot_rig) / cell_val) * 100
            elif col == ccl - 1:
                tot_rig = array_df[lin][lin]
                per_ok = (float(tot_rig) / cell_val) * 100
            elif lin == ccl - 1:
                tot_rig = array_df[col][col]
                per_ok = (float(tot_rig) / cell_val) * 100
            per_err = 100 - per_ok
        else:
            per_ok = per_err = 0

        per_ok_s = ["%.2f%%" % (per_ok), "100%"][per_ok == 100]

        # text to DEL
        text_del.append(oText)

        # text to ADD
        font_prop = fm.FontProperties(weight="bold", size=fz)
        text_kwargs = dict(
            color="w",
            ha="center",
            va="center",
            gid="sum",
            fontproperties=font_prop,
        )
        lis_txt = ["%d" % (cell_val), per_ok_s, "%.2f%%" % (per_err)]
        lis_kwa = [text_kwargs]
        dic = text_kwargs.copy()
        dic["color"] = "g"
        lis_kwa.append(dic)
        dic = text_kwargs.copy()
        dic["color"] = "r"
        lis_kwa.append(dic)
        lis_pos = [
            (oText._x, oText._y - 0.3),
            (oText._x, oText._y),
            (oText._x, oText._y + 0.3),
        ]
        for i in range(len(lis_txt)):
            newText = dict(
                x=lis_pos[i][0],
                y=lis_pos[i][1],
                text=lis_txt[i],
                kw=lis_kwa[i],
            )
            text_add.append(newText)

        # set background color for sum cells (last line and last column)
        carr = [0.27, 0.30, 0.27, 1.0]
        if (col == ccl - 1) and (lin == ccl - 1):
            carr = [0.17, 0.20, 0.17, 1.0]
        facecolors[posi] = carr

    else:
        if per > 0:
            txt = "%s\n%.2f%%" % (cell_val, per)
        else:
            if show_null_values == 0:
                txt = ""
            elif show_null_values == 1:
                txt = "0"
            else:
                txt = "0\n0.0%"
        oText.set_text(txt)

        # main diagonal
        if col == lin:
            # set color of the textin the diagonal to white
            oText.set_color("w")
            # set background color in the diagonal to blue
            facecolors[posi] = [0.35, 0.8, 0.55, 1.0]
        else:
            oText.set_color("r")

    return text_add, text_del

##################################################################################################################

def insert_totals(df_cm):
    """insert total column and line (the last ones)"""
    sum_col = []
    for c in df_cm.columns:
        sum_col.append(df_cm[c].sum())
    sum_lin = []
    for item_line in df_cm.iterrows():
        sum_lin.append(item_line[1].sum())
    df_cm["sum_lin"] = sum_lin
    sum_col.append(np.sum(sum_lin))
    df_cm.loc["sum_col"] = sum_col

##################################################################################################################

def pp_matrix(df_cm, annot=True,cmap="Oranges", fmt=".2f",fz=11, lw=0.5, cbar=False,figsize=[8, 8],show_null_values=0,pred_val_axis="y"):
    if pred_val_axis in ("col", "x"):
        xlbl = "Predicted"
        ylbl = "Actual"
    else:
        xlbl = "Actual"
        ylbl = "Predicted"
        df_cm = df_cm.T

    # create "Total" column
    insert_totals(df_cm)

    # this is for print allways in the same window
    fig, ax1 = get_new_fig("Conf matrix default", figsize)

    ax = sn.heatmap( df_cm,annot=annot,annot_kws={"size": fz}, linewidths=lw, ax=ax1, cbar=cbar, cmap=cmap, linecolor="w",fmt=fmt,)

    # set ticklabels rotation
    ax.set_xticklabels(ax.get_xticklabels(), rotation=45, fontsize=10)
    ax.set_yticklabels(ax.get_yticklabels(), rotation=25, fontsize=10)

    # Turn off all the ticks
    for t in ax.xaxis.get_major_ticks():
        t.tick1On = False
        t.tick2On = False
    for t in ax.yaxis.get_major_ticks():
        t.tick1On = False
        t.tick2On = False

    # face colors list
    quadmesh = ax.findobj(QuadMesh)[0]
    facecolors = quadmesh.get_facecolors()

    # iter in text elements
    array_df = np.array(df_cm.to_records(index=False).tolist())
    text_add = []
    text_del = []
    posi = -1  # from left to right, bottom to top.
    for t in ax.collections[0].axes.texts:  # ax.texts:
        pos = np.array(t.get_position()) - [0.5, 0.5]
        lin = int(pos[1])
        col = int(pos[0])
        posi += 1

        # set text
        txt_res = configcell_text_and_colors(
            array_df, lin, col, t, facecolors, posi, fz, fmt, show_null_values
        )

        text_add.extend(txt_res[0])
        text_del.extend(txt_res[1])

    # remove the old ones
    for item in text_del:
        item.remove()
    # append the new ones
    for item in text_add:
        ax.text(item["x"], item["y"], item["text"], **item["kw"])

    # titles and legends
    ax.set_title("Confusion matrix")
    ax.set_xlabel(xlbl)
    ax.set_ylabel(ylbl)
    plt.tight_layout()  # set layout slim
    plt.savefig(folder_name+"/com.png")
    plt.show()

##################################################################################################################

def pp_matrix_from_data(y_test,predictions, columns=None, annot=True,cmap="Oranges",fmt=".2f",fz=11, lw=0.5, cbar=False,figsize=[8, 8],show_null_values=0, pred_val_axis="lin",):
    if not columns:
        from string import ascii_uppercase

        columns = [
            "class %s" % (i)
            for i in list(ascii_uppercase)[0 : len(np.unique(y_test))]
        ]

    confm = sklearn.metrics.confusion_matrix(y_test, predictions)
    df_cm = pd.DataFrame(confm, index=columns, columns=columns)
    pp_matrix(df_cm, fz=fz,cmap=cmap, figsize=figsize, show_null_values=show_null_values, pred_val_axis=pred_val_axis,)


##################################################################################################################

folder_experiment_name="exp/"
train_batch_size =  1024
test_batch_size =  1000
epoch_numbers = 20

experiment_name = folder_experiment_name+"final_vgg16_cifar10_triplet_sgdmomentum_layer_nofreeze_adaptivelrsi_earlystopping_22222"+str(train_batch_size)

# get the current date and time
now = datetime.datetime.now()
str_time = str(now).replace(" ","_")
file_name = experiment_name+"_"+str_time+"/output.txt"
writer = SummaryWriter(file_name.replace(".txt",""))
file_experiment = open(file_name, "a")
file_experiment.write("date time : "+str(datetime.datetime.now())+"\n")
file_experiment.write("train batch size : "+str(train_batch_size)+"\n")
file_experiment.write("test batch size : "+str(test_batch_size)+"\n")
file_experiment.write("epoch number : "+str(epoch_numbers)+"\n")
print(file_name)

transform_train = transforms.Compose([
    transforms.RandomCrop(32, padding=4),
    transforms.RandomHorizontalFlip(),
    transforms.ToTensor(),
    transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
])

transform_test = transforms.Compose([
    transforms.ToTensor(),
   transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
])
train_set = CIFAR10(".",train=True,download=True,transform=transform_train,target_transform=lambda y: torch.tensor(y))
val_set = CIFAR10(".",train=False,download=True,transform=transform_test,target_transform=lambda y: torch.tensor(y))
train_loader = DataLoader(train_set,batch_size=train_batch_size,shuffle=True)
val_loader = DataLoader(val_set,batch_size=test_batch_size,shuffle=False)


model = ConvBlock(in_channels=3,layer_map=[64]
                  ,in_shape=[3,32,32]
                 )
distance = distances.CosineSimilarity()
reducer = reducers.ThresholdReducer(low=0)
criterion = losses.TripletMarginLoss(margin=0.2, distance=distance, reducer=reducer)
mining_func = miners.TripletMarginMiner(
    margin=0.2, distance=distance, type_of_triplets="semihard"
)



def train(epochs,count_layer,file_experiment):
    p = 0
    best_si_layer=0
    count_no_chagne_si=0
    Saved_si=0
    si_train=[]
    loop = trange(epochs)
    count_epoch = 0
    for epoch in loop:
        r_loss = 0
        r_acc = 0
        count_epoch=count_epoch+1
        training_loop = tqdm(train_loader,leave=False)
        training_loop.set_description("training")
        model.train()
        for X,y in training_loop:
            y = y.to(device)
            out = model(X.to(device))
            indices_tuple = mining_func(out, y)
            loss = criterion(out,y,indices_tuple)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            r_loss+=loss.item()
            r_acc+=accuracy(out,y)
            
        logs["training loss"].append(r_loss/len(train_loader))
        logs["training accuracy"].append(r_acc/len(train_loader))
        writer.add_scalar('training loss/'+str(count_layer),r_loss/len(train_loader),count_epoch)
        writer.add_scalar('training accuracy/'+str(count_layer),r_acc/len(train_loader),count_epoch)
        
        
        
        r_loss = 0
        r_acc = 0
       
        val_loop = tqdm(val_loader,leave=False)
        val_loop.set_description("validation")
        model.eval()

         #**********************************************************************************
#calculation of si 

        lens = [] 
        y_cpu = y.detach().cpu().numpy()
        for i in range(10):
            l=len(np.where(y_cpu==i)[0])
            lens.append(l)
#             print(f"class {i}: {l}")

                
        out_cpu=out.detach().cpu().numpy()
        
        separation = 0
        centers = np.zeros((10,out_cpu.shape[1]))
        
        for feat,label in zip(out_cpu,y_cpu):
           
            centers[label,:]= centers[label,:] +  feat

        for i,l in enumerate(lens):
              centers[i,:]/=l


        for feat,true_label in zip(out_cpu,y_cpu):
            min_dist = 10000
            center_label= 10000
            for center,l in zip(centers,range(10)):
                dist = np.linalg.norm(feat-center)
                if dist < min_dist:
                    min_dist = dist
                    center_label = l
            if true_label == center_label:
                separation+=1 
        si_value=float(separation/out_cpu.shape[0])
        if(si_value > best_si_layer):
            best_si_layer=si_value
            count_no_chagne_si=0
        else:
            count_no_chagne_si=count_no_chagne_si+1
        if(count_no_chagne_si>=5):
            break
        if(count_epoch == 1):
            Saved_si=si_value
            print_si_train="SI_train epoch (min) ("+str(count_epoch)+") : " + str(Saved_si) 
        else:
            print_si_train="SI_train epoch ("+str(count_epoch)+") : " + str(si_value) 
        writer.add_scalar('SI/'+str(count_layer),si_value,count_epoch)
        writer.add_scalar('LR/'+str(count_layer),scheduler.get_last_lr(),count_epoch)
        file_experiment.write(print_si_train)
        print(print_si_train)
        SI_Center=separation/out_cpu.shape[0]
        si_train.append(SI_Center)
        #################################### validation #######################################################
        for X,y in val_loop:
            y = y.to(device)
            with torch.no_grad():
                out = model(X.to(device))
                indices_tuple = mining_func(out, y)
                loss = criterion(out,y,indices_tuple)
                r_loss+=loss.item()
                r_acc+=accuracy(out,y)
        
        
        logs["val loss"].append(r_loss/len(val_loader))
        logs["val accuracy"].append(r_acc/len(val_loader))
        writer.add_scalar('val loss/'+str(count_layer),r_loss/len(val_loader),count_epoch)
        writer.add_scalar('val accuracy/'+str(count_layer),r_acc/len(val_loader),count_epoch)

        
        
        loop.set_postfix({"train loss":logs["training loss"][-1],"train accuracy":logs["training accuracy"][-1],
                          "val loss":logs["val loss"][-1],"val accuracy":logs["val accuracy"][-1]})

        
        scheduler.step()
    
    print_train_loss = "Train loss layer ("+str(count_layer)+") : " + str(logs["training loss"][-1]) +"\n"
    print_val_loss ="Val loss layer ("+str(count_layer)+") : " + str(logs["val loss"][-1]) +"\n"
    file_experiment.write(print_train_loss) 
    file_experiment.write(print_val_loss)
    file_experiment.close()
    return Saved_si,best_si_layer



#######################################################################################################################


file_experiment.close()
file_experiment = open(file_name, "a")
logs_total = []
count_layer = 0
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


logs = {"training loss":[],"training accuracy":[],"val loss":[],"val accuracy":[]}
model = ConvBlock(in_channels=3,layer_map=[64]
                  ,in_shape=[3,32,32]
                 ).to(device)
optimizer = torch.optim.SGD(model.parameters(),lr=0.001,momentum=0.9) 
scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer,gamma=0.95)
Saved_si,best_si_layer = train(epoch_numbers,count_layer,file_experiment)
model = model.cpu()
logs_total.append(logs)




in_channels = 64
layer_list = [[64, 'm'], [128], [128, 'm'], [256], [256], [256, 'm'], [512], [512], [512, 'm'], [512], [512],[ 512, 'm']]
si_count = -1
for layer in tqdm(layer_list):
    count_layer = count_layer+1
    si_count =  si_count+1
    file_experiment.close()
    file_experiment = open(file_name, "a")
    logs = {"training loss":[],"training accuracy":[],"val loss":[],"val accuracy":[]}
    model = ModuleAdder(model.conv_layers,ConvBlock(in_channels=in_channels,layer_map=layer),freeze=False).cuda()
    final_learning_rate=Saved_si*0.03
    print_layer_learning_rate = "Learning rate layer ("+str(count_layer)+") : "+ str(final_learning_rate)+"\n"
    writer.add_scalar('LR_Layer',final_learning_rate,count_layer)
    file_experiment.write(print_layer_learning_rate)
    print(print_layer_learning_rate)
    
    optimizer = torch.optim.SGD(model.parameters(),lr=final_learning_rate,momentum=0.9) 
    scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer,gamma=0.95)
    logs_total.append(logs)
    in_channels = layer[0]
    Saved_si,best_si_layer = train(epoch_numbers,count_layer,file_experiment)
    writer.add_scalar('SI_Layer',best_si_layer,count_layer)
    model = model.cpu()

##################################################################################################################################################

torch.save(model,file_name.replace(".txt","_conv.pt"))
file_experiment = open(file_name, "a")
file_experiment.write(str(logs_total))
file_experiment.close()


model = MLP(model.conv_layers).cuda()
init_lr=0.01
optimizer = torch.optim.SGD(model.parameters(),lr=init_lr,momentum=0.9) 


##################################################################################################################################################



device = "cuda" if torch.cuda.is_available() else "cpu"
criterion_cce = nn.CrossEntropyLoss()
logs = {"training loss":[],"training accuracy":[],"validation loss":[],"validation accuracy":[]}

def train_cce(epochs,file_experiment):
    p = 0
    count_epoch=0
    loop = trange(epochs)
    best_val_acc=0
    best_y_pred=[]
    best_y_test=[]
    for epoch in loop:
        count_epoch=count_epoch+1
        r_loss = 0
        r_acc = 0
        training_loop = tqdm(train_loader,leave=False)
        training_loop.set_description("training")
        model.train()
        for X,y in training_loop:
            y = y.to(device)
            out = model(X.to(device))
            loss = criterion_cce(out,y.to(device))
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            r_loss+=loss.item()
            r_acc+=accuracy(out,y)
            
            
            
        train_loss=r_loss/len(train_loader)
        train_acc=r_acc/len(train_loader)
        logs["training accuracy"].append(train_acc)
        logs["training loss"].append(train_loss)
        
        writer.add_scalar('train loss/'+str(count_layer),train_loss,count_epoch)
        writer.add_scalar('train accuracy/'+str(count_layer),train_acc,count_epoch)
        
        
        
        r_loss = 0
        r_acc = 0
        val_loop = tqdm(val_loader,leave=False)
        val_loop.set_description("validation")
        model.eval()
        first_validation=0
        for X,y in val_loop:
            y = y.to(device)
            with torch.no_grad():
                out = model(X.to(device))
                loss = criterion_cce(out,y)

                r_loss+=loss.item()
                r_acc+=accuracy(out,y)
                out_predict=torch.argmax(out,dim=1)
                if(first_validation == 0):
                    y_pred=out_predict.cpu().numpy()
                    y_test =y.cpu().numpy()
                else:
                    y_pred = np.append(y_pred,out_predict.cpu().numpy())
                    y_test = np.append(y_test,y.cpu().numpy())
                first_validation=1
                
        val_loss=r_loss/len(val_loader)
        val_acc=r_acc/len(val_loader)
        if(val_acc>best_val_acc):
            best_val_acc=val_acc
            best_y_pred=y_pred
            best_y_test=y_test
        logs["validation accuracy"].append(val_acc)
        logs["validation loss"].append(val_loss)
        
        writer.add_scalar('valid loss/'+str(count_layer),val_loss,count_epoch)
        writer.add_scalar('valid accuracy/'+str(count_layer),val_acc,count_epoch)
        
        
        
        
        loop.set_postfix({"train loss":logs["training loss"][-1],"val loss":logs["validation loss"][-1],
                         "train acc":logs["training accuracy"][-1],"val acc":logs["validation accuracy"][-1]})
        
        scheduler.step()
        file_experiment.write("train loss epoch ("+str(count_epoch)+") : "+ str(logs["training loss"][-1])+"\n")
        file_experiment.write("validation loss epoch ("+str(count_epoch)+") : "+ str(logs["validation loss"][-1])+"\n")
        file_experiment.write("training accuracy epoch ("+str(count_epoch)+") : "+ str(logs["training accuracy"][-1])+"\n")
        file_experiment.write("validation accuracy epoch ("+str(count_epoch)+") : "+ str(logs["validation accuracy"][-1])+"\n")
        
        # if logs["training accuracy"][-1]>logs["validation accuracy"][-1]:
        # #     p+=1
        # # if p>2:
        #     return
    file_experiment.close()
    return best_y_pred,best_y_test



##################################################################################################################################################

file_experiment = open(file_name, "a")
file_experiment.write("train_cce \n")
y_pred,y_test = train_cce(25,file_experiment)
file_experiment.close()

##################################################################################################################################################

folder_name=experiment_name+"_"+str_time
pp_matrix_from_data(y_test, y_pred, cmap='PuRd')

##################################################################################################################################################

df = pd.DataFrame(sklearn.metrics.classification_report(y_pred,  y_test, digits=2,output_dict=True)).T
df['support'] = df.support.apply(int)
df.style.background_gradient(cmap='viridis',subset=pd.IndexSlice['0':'9', :'f1-score']).to_excel(folder_name+"/classification.xlsx", engine='openpyxl')

